@@ -1484,13 +1484,9 @@ static void http_cache_io_handler(struct appctx *appctx)
 	if (unlikely(se_fl_test(appctx->sedesc, (SE_FL_EOS|SE_FL_ERROR|SE_FL_SHR|SE_FL_SHW))))
 		goto out;
 
-	/* Check if the input buffer is available. */
-	if (!b_size(&res->buf)) {
-		sc_need_room(sc, 0);
-		goto out;
-	}
-
 	if (appctx->st0 == HTX_CACHE_INIT) {
+		if (!co_data(req))
+			goto wait_request;
 		ctx->next = block_ptr(cache_ptr);
 		ctx->offset = sizeof(*cache_ptr);
 		ctx->sent = 0;
@@ -1498,6 +1494,12 @@ static void http_cache_io_handler(struct appctx *appctx)
 		appctx->st0 = HTX_CACHE_HEADER;
 	}
 
+	/* Check if the input buffer is available. */
+	if (!b_size(&res->buf)) {
+		sc_need_room(sc, 0);
+		goto out;
+	}
+
 	if (appctx->st0 == HTX_CACHE_HEADER) {
 		/* Headers must be dump at once. Otherwise it is an error */
 		len = first->len - sizeof(*cache_ptr) - ctx->sent;
@@ -1560,6 +1562,12 @@ static void http_cache_io_handler(struct appctx *appctx)
 	}
 	return;
 
+  wait_request:
+	/* Wait for the request before starting to deliver the response */
+	b_reset(&res->buf);
+	applet_need_more_data(appctx);
+	return;
+
   error:
 	/* Sent and HTTP error 500 */
 	b_reset(&res->buf);
