@@ -33,7 +33,8 @@
 #include <haproxy/pool.h>
 #include <haproxy/regex.h>
 #include <haproxy/sample.h>
-#include <haproxy/stream_interface.h>
+#include <haproxy/sc_strm.h>
+#include <haproxy/stconn.h>
 #include <haproxy/tools.h>
 #include <haproxy/uri_auth-t.h>
 #include <haproxy/uri_normalizer.h>
@@ -178,6 +179,7 @@ static enum act_parse_ret parse_set_req_line(const char **args, int *orig_arg, s
 	}
 	rule->action_ptr = http_action_set_req_line;
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	if (!*args[cur_arg] ||
 	    (*args[cur_arg + 1] && strcmp(args[cur_arg + 1], "if") != 0 && strcmp(args[cur_arg + 1], "unless") != 0)) {
@@ -185,7 +187,6 @@ static enum act_parse_ret parse_set_req_line(const char **args, int *orig_arg, s
 		return ACT_RET_PRS_ERR;
 	}
 
-	LIST_INIT(&rule->arg.http.fmt);
 	px->conf.args.ctx = ARGC_HRQ;
 	if (px->cap & PR_CAP_FE)
 		cap |= SMP_VAL_FE_HRQ_HDR;
@@ -615,6 +616,7 @@ static enum act_parse_ret parse_replace_uri(const char **args, int *orig_arg, st
 
 	rule->action_ptr = http_action_replace_uri;
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	if (!*args[cur_arg] || !*args[cur_arg+1] ||
 	    (*args[cur_arg+2] && strcmp(args[cur_arg+2], "if") != 0 && strcmp(args[cur_arg+2], "unless") != 0)) {
@@ -628,7 +630,6 @@ static enum act_parse_ret parse_replace_uri(const char **args, int *orig_arg, st
 		return ACT_RET_PRS_ERR;
 	}
 
-	LIST_INIT(&rule->arg.http.fmt);
 	px->conf.args.ctx = ARGC_HRQ;
 	if (px->cap & PR_CAP_FE)
 		cap |= SMP_VAL_FE_HRQ_HDR;
@@ -679,6 +680,7 @@ static enum act_parse_ret parse_http_set_status(const char **args, int *orig_arg
 	rule->action = ACT_CUSTOM;
 	rule->action_ptr = action_http_set_status;
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	/* Check if an argument is available */
 	if (!*args[*orig_arg]) {
@@ -700,12 +702,10 @@ static enum act_parse_ret parse_http_set_status(const char **args, int *orig_arg
 	if (*args[*orig_arg] && strcmp(args[*orig_arg], "reason") == 0 &&
 	    (*args[*orig_arg + 1] && strcmp(args[*orig_arg + 1], "if") != 0 && strcmp(args[*orig_arg + 1], "unless") != 0)) {
 		(*orig_arg)++;
-		rule->arg.http.str.ptr = strdup(args[*orig_arg]);
-		rule->arg.http.str.len = strlen(rule->arg.http.str.ptr);
+		rule->arg.http.str = ist(strdup(args[*orig_arg]));
 		(*orig_arg)++;
 	}
 
-	LIST_INIT(&rule->arg.http.fmt);
 	return ACT_RET_PRS_OK;
 }
 
@@ -719,9 +719,8 @@ static enum act_parse_ret parse_http_set_status(const char **args, int *orig_arg
 static enum act_return http_action_reject(struct act_rule *rule, struct proxy *px,
                                           struct session *sess, struct stream *s, int flags)
 {
-	si_must_kill_conn(chn_prod(&s->req));
-	channel_abort(&s->req);
-	channel_abort(&s->res);
+	sc_must_kill_conn(s->scf);
+	stream_abort(s);
 	s->req.analysers &= AN_REQ_FLT_END;
 	s->res.analysers &= AN_RES_FLT_END;
 
@@ -753,19 +752,17 @@ static enum act_parse_ret parse_http_action_reject(const char **args, int *orig_
 /* This function executes the "disable-l7-retry" HTTP action.
  * It disables L7 retries (all retry except for a connection failure). This
  * can be useful for example to avoid retrying on POST requests.
- * It just removes the L7 retry flag on the stream_interface, and always
+ * It just removes the L7 retry flag on the HTTP transaction, and always
  * return ACT_RET_CONT;
  */
 static enum act_return http_req_disable_l7_retry(struct act_rule *rule, struct proxy *px,
                                           struct session *sess, struct stream *s, int flags)
 {
-	struct stream_interface *si = &s->si[1];
-
-	/* In theory, the SI_FL_L7_RETRY flags isn't set at this point, but
+	/* In theory, the TX_L7_RETRY flags isn't set at this point, but
 	 * let's be future-proof and remove it anyway.
 	 */
-	si->flags &= ~SI_FL_L7_RETRY;
-	si->flags |= SI_FL_D_L7_RETRY;
+	s->txn->flags &= ~TX_L7_RETRY;
+	s->txn->flags |= TX_D_L7_RETRY;
 	return ACT_RET_CONT;
 }
 
@@ -1287,7 +1284,7 @@ static enum act_return http_action_auth(struct act_rule *rule, struct proxy *px,
 		goto fail;
 
 	/* Note: Only eval on the request */
-	s->logs.tv_request = now;
+	s->logs.request_ts = now_ns;
 	req->analysers &= AN_REQ_FLT_END;
 
 	if (s->sess->fe == s->be) /* report it if the request was intercepted by the frontend */
@@ -1320,6 +1317,7 @@ static enum act_parse_ret parse_http_auth(const char **args, int *orig_arg, stru
 	rule->flags |= ACT_FLAG_FINAL;
 	rule->action_ptr = http_action_auth;
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	cur_arg = *orig_arg;
 	if (strcmp(args[cur_arg], "realm") == 0) {
@@ -1328,12 +1326,10 @@ static enum act_parse_ret parse_http_auth(const char **args, int *orig_arg, stru
 			memprintf(err, "missing realm value.\n");
 			return ACT_RET_PRS_ERR;
 		}
-		rule->arg.http.str.ptr = strdup(args[cur_arg]);
-		rule->arg.http.str.len = strlen(rule->arg.http.str.ptr);
+		rule->arg.http.str = ist(strdup(args[cur_arg]));
 		cur_arg++;
 	}
 
-	LIST_INIT(&rule->arg.http.fmt);
 	*orig_arg = cur_arg;
 	return ACT_RET_PRS_OK;
 }
@@ -1382,7 +1378,7 @@ static enum act_return http_action_early_hint(struct act_rule *rule, struct prox
 		s->txn->status = 103;
 	}
 
-	/* Add the HTTP Early Hint HTTP 103 response heade */
+	/* Add the HTTP Early Hint HTTP 103 response header */
 	value->data = build_logline(s, b_tail(value), b_room(value), &rule->arg.http.fmt);
 	if (!htx_add_header(htx, rule->arg.http.str, ist2(b_head(value), b_data(value))))
 		goto error;
@@ -1487,6 +1483,7 @@ static enum act_parse_ret parse_http_set_header(const char **args, int *orig_arg
 						   struct act_rule *rule, char **err)
 {
 	int cap = 0, cur_arg;
+	const char *p;
 
 	if (args[*orig_arg-1][0] == 'e') {
 		rule->action = ACT_CUSTOM;
@@ -1500,6 +1497,7 @@ static enum act_parse_ret parse_http_set_header(const char **args, int *orig_arg
 		rule->action_ptr = http_action_set_header;
 	}
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	cur_arg = *orig_arg;
 	if (!*args[cur_arg] || !*args[cur_arg+1]) {
@@ -1508,9 +1506,7 @@ static enum act_parse_ret parse_http_set_header(const char **args, int *orig_arg
 	}
 
 
-	rule->arg.http.str.ptr = strdup(args[cur_arg]);
-	rule->arg.http.str.len = strlen(rule->arg.http.str.ptr);
-	LIST_INIT(&rule->arg.http.fmt);
+	rule->arg.http.str = ist(strdup(args[cur_arg]));
 
 	if (rule->from == ACT_F_HTTP_REQ) {
 		px->conf.args.ctx = ARGC_HRQ;
@@ -1537,6 +1533,23 @@ static enum act_parse_ret parse_http_set_header(const char **args, int *orig_arg
 	px->conf.lfs_file = strdup(px->conf.args.file);
 	px->conf.lfs_line = px->conf.args.line;
 
+	/* some characters are totally forbidden in header names and
+	 * may happen by accident when writing configs, causing strange
+	 * failures in field. Better catch these ones early, nobody will
+	 * miss them. In particular, a colon at the end (or anywhere
+	 * after the first char) or a space/cr anywhere due to misplaced
+	 * quotes are hard to spot.
+	 */
+	for (p = istptr(rule->arg.http.str); p < istend(rule->arg.http.str); p++) {
+		if (HTTP_IS_TOKEN(*p))
+			continue;
+		if (p == istptr(rule->arg.http.str) && *p == ':')
+			continue;
+		/* we only report this as-is but it will not cause an error */
+		memprintf(err, "header name '%s' contains forbidden character '%c'", istptr(rule->arg.http.str), *p);
+		break;
+	}
+
 	*orig_arg = cur_arg + 1;
 	return ACT_RET_PRS_OK;
 }
@@ -1610,6 +1623,7 @@ static enum act_parse_ret parse_http_replace_header(const char **args, int *orig
 		rule->action = 1; // replace-value
 	rule->action_ptr = http_action_replace_header;
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	cur_arg = *orig_arg;
 	if (!*args[cur_arg] || !*args[cur_arg+1] || !*args[cur_arg+2]) {
@@ -1617,9 +1631,7 @@ static enum act_parse_ret parse_http_replace_header(const char **args, int *orig
 		return ACT_RET_PRS_ERR;
 	}
 
-	rule->arg.http.str.ptr = strdup(args[cur_arg]);
-	rule->arg.http.str.len = strlen(rule->arg.http.str.ptr);
-	LIST_INIT(&rule->arg.http.fmt);
+	rule->arg.http.str = ist(strdup(args[cur_arg]));
 
 	cur_arg++;
 	if (!(rule->arg.http.re = regex_comp(args[cur_arg], 1, 1, err))) {
@@ -1714,6 +1726,7 @@ static enum act_parse_ret parse_http_del_header(const char **args, int *orig_arg
 	rule->action = PAT_MATCH_STR;
 	rule->action_ptr = http_action_del_header;
 	rule->release_ptr = release_http_action;
+	LIST_INIT(&rule->arg.http.fmt);
 
 	cur_arg = *orig_arg;
 	if (!*args[cur_arg]) {
@@ -1721,11 +1734,9 @@ static enum act_parse_ret parse_http_del_header(const char **args, int *orig_arg
 		return ACT_RET_PRS_ERR;
 	}
 
-	rule->arg.http.str.ptr = strdup(args[cur_arg]);
-	rule->arg.http.str.len = strlen(rule->arg.http.str.ptr);
+	rule->arg.http.str = ist(strdup(args[cur_arg]));
 	px->conf.args.ctx = (rule->from == ACT_F_HTTP_REQ ? ARGC_HRQ : ARGC_HRS);
 
-	LIST_INIT(&rule->arg.http.fmt);
 	if (strcmp(args[cur_arg+1], "-m") == 0) {
 		cur_arg++;
 		if (!*args[cur_arg+1]) {
@@ -1739,7 +1750,7 @@ static enum act_parse_ret parse_http_del_header(const char **args, int *orig_arg
 		case PAT_MATCH_REG:
 			if (!(rule->arg.http.re = regex_comp(rule->arg.http.str.ptr, 1, 1, err)))
 				return ACT_RET_PRS_ERR;
-			/* fall through */
+			__fallthrough;
 		case PAT_MATCH_STR:
 		case PAT_MATCH_BEG:
 		case PAT_MATCH_END:
@@ -1759,24 +1770,14 @@ static enum act_parse_ret parse_http_del_header(const char **args, int *orig_arg
 /* Release memory allocated by an http redirect action. */
 static void release_http_redir(struct act_rule *rule)
 {
-	struct logformat_node *lf, *lfb;
 	struct redirect_rule *redir;
 
 	redir = rule->arg.redir;
+	if (!redir)
+		return;
+
 	LIST_DELETE(&redir->list);
-	if (redir->cond) {
-		prune_acl_cond(redir->cond);
-		free(redir->cond);
-	}
-	free(redir->rdr_str);
-	free(redir->cookie_str);
-	list_for_each_entry_safe(lf, lfb, &redir->rdr_fmt, list) {
-		LIST_DELETE(&lf->list);
-		release_sample_expr(lf->expr);
-		free(lf->arg);
-		free(lf);
-	}
-	free(redir);
+	http_free_redirect_rule(redir);
 }
 
 /* Parse a "redirect" action. It returns ACT_RET_PRS_OK on success,
@@ -2265,7 +2266,7 @@ static enum act_return http_action_return(struct act_rule *rule, struct proxy *p
 
 	if (rule->from == ACT_F_HTTP_REQ) {
 		/* let's log the request time */
-		s->logs.tv_request = now;
+		s->logs.request_ts = now_ns;
 		req->analysers &= AN_REQ_FLT_END;
 
 		if (s->sess->fe == s->be) /* report it if the request was intercepted by the frontend */
@@ -2481,10 +2482,13 @@ static struct action_kw_list http_after_res_actions = {
 		{ "add-header",      parse_http_set_header,     0 },
 		{ "allow",           parse_http_allow,          0 },
 		{ "capture",         parse_http_res_capture,    0 },
+		{ "del-acl",          parse_http_set_map,       KWF_MATCH_PREFIX },
 		{ "del-header",      parse_http_del_header,     0 },
+		{ "del-map",          parse_http_set_map,       KWF_MATCH_PREFIX },
 		{ "replace-header",  parse_http_replace_header, 0 },
 		{ "replace-value",   parse_http_replace_header, 0 },
 		{ "set-header",      parse_http_set_header,     0 },
+		{ "set-map",         parse_http_set_map,        KWF_MATCH_PREFIX },
 		{ "set-status",      parse_http_set_status,     0 },
 		{ "strict-mode",     parse_http_strict_mode,    0 },
 		{ NULL, NULL }
