@@ -18,8 +18,13 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include <sys/mman.h>
+#include <errno.h>
+#include <fcntl.h>
+
 #include <import/ist.h>
 #include <haproxy/api.h>
+#include <haproxy/applet.h>
 #include <haproxy/cfgparse.h>
 #include <haproxy/cli.h>
 #include <haproxy/errors.h>
@@ -27,9 +32,10 @@
 #include <haproxy/log.h>
 #include <haproxy/proxy.h>
 #include <haproxy/ring.h>
+#include <haproxy/sc_strm.h>
 #include <haproxy/signal.h>
 #include <haproxy/sink.h>
-#include <haproxy/stream_interface.h>
+#include <haproxy/stconn.h>
 #include <haproxy/time.h>
 #include <haproxy/tools.h>
 
@@ -162,10 +168,13 @@ struct sink *sink_new_buf(const char *name, const char *desc, enum log_fmt fmt,
  * array <msg> to sink <sink>. Formatting according to the sink's preference is
  * done here. Lost messages are NOT accounted for. It is preferable to call
  * sink_write() instead which will also try to emit the number of dropped
- * messages when there are any. It returns >0 if it could write anything,
- * <=0 otherwise.
+ * messages when there are any. It will stop writing at <maxlen> instead of
+ * sink->maxlen if <maxlen> is positive and inferior to sink->maxlen.
+ *
+ * It returns >0 if it could write anything, <=0 otherwise.
  */
- ssize_t __sink_write(struct sink *sink, const struct ist msg[], size_t nmsg,
+ ssize_t __sink_write(struct sink *sink, size_t maxlen,
+	             const struct ist msg[], size_t nmsg,
 	             int level, int facility, struct ist *metadata)
  {
 	struct ist *pfx = NULL;
@@ -177,11 +186,13 @@ struct sink *sink_new_buf(const char *name, const char *desc, enum log_fmt fmt,
 	pfx = build_log_header(sink->fmt, level, facility, metadata, &npfx);
 
 send:
+	if (!maxlen)
+		maxlen = ~0;
 	if (sink->type == SINK_TYPE_FD) {
-		return fd_write_frag_line(sink->ctx.fd, sink->maxlen, pfx, npfx, msg, nmsg, 1);
+		return fd_write_frag_line(sink->ctx.fd, MIN(maxlen, sink->maxlen), pfx, npfx, msg, nmsg, 1);
 	}
 	else if (sink->type == SINK_TYPE_BUFFER) {
-		return ring_write(sink->ctx.ring, sink->maxlen, pfx, npfx, msg, nmsg);
+		return ring_write(sink->ctx.ring, MIN(maxlen, sink->maxlen), pfx, npfx, msg, nmsg);
 	}
 	return 0;
 }
@@ -208,7 +219,7 @@ int sink_announce_dropped(struct sink *sink, int facility)
 
 		if (!metadata[LOG_META_HOST].len) {
 			if (global.log_send_hostname)
-				metadata[LOG_META_HOST] = ist2(global.log_send_hostname, strlen(global.log_send_hostname));
+				metadata[LOG_META_HOST] = ist(global.log_send_hostname);
 		}
 
 		if (!metadata[LOG_META_TAG].len)
@@ -223,7 +234,7 @@ int sink_announce_dropped(struct sink *sink, int facility)
 			metadata[LOG_META_PID] = ist2(pidstr, strlen(pidstr));
 		}
 
-		if (__sink_write(sink, msgvec, 1, LOG_NOTICE, facility, metadata) <= 0)
+		if (__sink_write(sink, 0, msgvec, 1, LOG_NOTICE, facility, metadata) <= 0)
 			return 0;
 		/* success! */
 		HA_ATOMIC_SUB(&sink->ctx.dropped, dropped);
@@ -235,6 +246,7 @@ int sink_announce_dropped(struct sink *sink, int facility)
 static int cli_parse_show_events(char **args, char *payload, struct appctx *appctx, void *private)
 {
 	struct sink *sink;
+	uint ring_flags;
 	int arg;
 
 	args++; // make args[1] the 1st arg
@@ -265,23 +277,24 @@ static int cli_parse_show_events(char **args, char *payload, struct appctx *appc
 	if (sink->type != SINK_TYPE_BUFFER)
 		return cli_msg(appctx, LOG_NOTICE, "Nothing to report for this sink");
 
+	ring_flags = 0;
 	for (arg = 2; *args[arg]; arg++) {
 		if (strcmp(args[arg], "-w") == 0)
-			appctx->ctx.cli.i0 |= 1; // wait mode
+			ring_flags |= RING_WF_WAIT_MODE;
 		else if (strcmp(args[arg], "-n") == 0)
-			appctx->ctx.cli.i0 |= 2; // seek to new
+			ring_flags |= RING_WF_SEEK_NEW;
 		else if (strcmp(args[arg], "-nw") == 0 || strcmp(args[arg], "-wn") == 0)
-			appctx->ctx.cli.i0 |= 3; // seek to new + wait
+			ring_flags |= RING_WF_WAIT_MODE | RING_WF_SEEK_NEW;
 		else
 			return cli_err(appctx, "unknown option");
 	}
-	return ring_attach_cli(sink->ctx.ring, appctx);
+	return ring_attach_cli(sink->ctx.ring, appctx, ring_flags);
 }
 
 /* Pre-configures a ring proxy to emit connections */
 void sink_setup_proxy(struct proxy *px)
 {
-	px->last_change = now.tv_sec;
+	px->last_change = ns_to_sec(now_ns);
 	px->cap = PR_CAP_BE;
 	px->maxconn = 0;
 	px->conn_retries = 1;
@@ -295,48 +308,36 @@ void sink_setup_proxy(struct proxy *px)
 }
 
 /*
- * IO Handler to handle message push to syslog tcp server
+ * IO Handler to handle message push to syslog tcp server.
+ * It takes its context from appctx->svcctx.
  */
 static void sink_forward_io_handler(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
-	struct stream *s = si_strm(si);
+	struct stconn *sc = appctx_sc(appctx);
+	struct stream *s = __sc_strm(sc);
 	struct sink *sink = strm_fe(s)->parent;
-	struct sink_forward_target *sft = appctx->ctx.sft.ptr;
+	struct sink_forward_target *sft = appctx->svcctx;
 	struct ring *ring = sink->ctx.ring;
 	struct buffer *buf = &ring->buf;
 	uint64_t msg_len;
 	size_t len, cnt, ofs, last_ofs;
 	int ret = 0;
 
+	if (unlikely(se_fl_test(appctx->sedesc, (SE_FL_EOS|SE_FL_ERROR|SE_FL_SHR|SE_FL_SHW))))
+		goto out;
+
 	/* if stopping was requested, close immediately */
 	if (unlikely(stopping))
 		goto close;
 
-	/* for rex because it seems reset to timeout
-	 * and we don't want expire on this case
-	 * with a syslog server
-	 */
-	si_oc(si)->rex = TICK_ETERNITY;
-	/* rto should not change but it seems the case */
-	si_oc(si)->rto = TICK_ETERNITY;
-
-	/* an error was detected */
-	if (unlikely(si_ic(si)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
-		goto close;
-
-	/* con closed by server side */
-	if ((si_oc(si)->flags & CF_SHUTW))
-		goto close;
-
 	/* if the connection is not established, inform the stream that we want
 	 * to be notified whenever the connection completes.
 	 */
-	if (si_opposite(si)->state < SI_ST_EST) {
-		si_cant_get(si);
-		si_rx_conn_blk(si);
-		si_rx_endp_more(si);
-		return;
+	if (sc_opposite(sc)->state < SC_ST_EST) {
+		applet_need_more_data(appctx);
+		se_need_remote_conn(appctx->sedesc);
+		applet_have_more_data(appctx);
+		goto out;
 	}
 
 	HA_SPIN_LOCK(SFT_LOCK, &sft->lock);
@@ -344,7 +345,6 @@ static void sink_forward_io_handler(struct appctx *appctx)
 		HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 		goto close;
 	}
-	ofs = sft->ofs;
 
 	HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 	LIST_DEL_INIT(&appctx->wait_entry);
@@ -360,132 +360,116 @@ static void sink_forward_io_handler(struct appctx *appctx)
 	 * existing messages before grabbing a reference to a location. This
 	 * value cannot be produced after initialization.
 	 */
-	if (unlikely(ofs == ~0)) {
-		ofs = 0;
-
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
+	if (unlikely(sft->ofs == ~0)) {
+		sft->ofs = b_peek_ofs(buf, 0);
+		HA_ATOMIC_INC(b_orig(buf) + sft->ofs);
 	}
 
+	/* we were already there, adjust the offset to be relative to
+	 * the buffer's head and remove us from the counter.
+	 */
+	ofs = sft->ofs - b_head_ofs(buf);
+	if (sft->ofs < b_head_ofs(buf))
+		ofs += b_size(buf);
+	BUG_ON(ofs >= buf->size);
+	HA_ATOMIC_DEC(b_peek(buf, ofs));
+
 	/* in this loop, ofs always points to the counter byte that precedes
 	 * the message so that we can take our reference there if we have to
 	 * stop before the end (ret=0).
 	 */
-	if (si_opposite(si)->state == SI_ST_EST) {
-		/* we were already there, adjust the offset to be relative to
-		 * the buffer's head and remove us from the counter.
-		 */
-		ofs -= ring->ofs;
-		BUG_ON(ofs >= buf->size);
-		HA_ATOMIC_DEC(b_peek(buf, ofs));
+	ret = 1;
+	while (ofs + 1 < b_data(buf)) {
+		cnt = 1;
+		len = b_peek_varint(buf, ofs + cnt, &msg_len);
+		if (!len)
+			break;
+		cnt += len;
+		BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
 
-		ret = 1;
-		while (ofs + 1 < b_data(buf)) {
-			cnt = 1;
-			len = b_peek_varint(buf, ofs + cnt, &msg_len);
-			if (!len)
-				break;
-			cnt += len;
-			BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
-
-			if (unlikely(msg_len + 1 > b_size(&trash))) {
-				/* too large a message to ever fit, let's skip it */
-				ofs += cnt + msg_len;
-				continue;
-			}
-
-			chunk_reset(&trash);
-			len = b_getblk(buf, trash.area, msg_len, ofs + cnt);
-			trash.data += len;
-			trash.area[trash.data++] = '\n';
-
-			if (ci_putchk(si_ic(si), &trash) == -1) {
-				si_rx_room_blk(si);
-				ret = 0;
-				break;
-			}
+		if (unlikely(msg_len + 1 > b_size(&trash))) {
+			/* too large a message to ever fit, let's skip it */
 			ofs += cnt + msg_len;
+			continue;
 		}
 
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
-		sft->ofs = ofs;
-		last_ofs = ring->ofs;
+		chunk_reset(&trash);
+		len = b_getblk(buf, trash.area, msg_len, ofs + cnt);
+		trash.data += len;
+		trash.area[trash.data++] = '\n';
+
+		if (applet_putchk(appctx, &trash) == -1) {
+			ret = 0;
+			break;
+		}
+		ofs += cnt + msg_len;
 	}
+
+	HA_ATOMIC_INC(b_peek(buf, ofs));
+	last_ofs = b_tail_ofs(buf);
+	sft->ofs = b_peek_ofs(buf, ofs);
+
 	HA_RWLOCK_RDUNLOCK(LOGSRV_LOCK, &ring->lock);
 
 	if (ret) {
 		/* let's be woken up once new data arrive */
 		HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 		LIST_APPEND(&ring->waiters, &appctx->wait_entry);
-		ofs = ring->ofs;
+		ofs = b_tail_ofs(buf);
 		HA_RWLOCK_WRUNLOCK(LOGSRV_LOCK, &ring->lock);
 		if (ofs != last_ofs) {
 			/* more data was added into the ring between the
 			 * unlock and the lock, and the writer might not
 			 * have seen us. We need to reschedule a read.
 			 */
-			si_rx_endp_more(si);
+			applet_have_more_data(appctx);
 		} else
-			si_rx_endp_done(si);
+			applet_have_no_more_data(appctx);
 	}
 	HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 
+out:
 	/* always drain data from server */
-	co_skip(si_oc(si), si_oc(si)->output);
+	co_skip(sc_oc(sc), sc_oc(sc)->output);
 	return;
 
 close:
-	si_shutw(si);
-	si_shutr(si);
-	si_ic(si)->flags |= CF_READ_NULL;
+	se_fl_set(appctx->sedesc, SE_FL_EOS|SE_FL_EOI);
 }
 
 /*
  * IO Handler to handle message push to syslog tcp server
  * using octet counting frames
+ * It takes its context from appctx->svcctx.
  */
 static void sink_forward_oc_io_handler(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
-	struct stream *s = si_strm(si);
+	struct stconn *sc = appctx_sc(appctx);
+	struct stream *s = __sc_strm(sc);
 	struct sink *sink = strm_fe(s)->parent;
-	struct sink_forward_target *sft = appctx->ctx.sft.ptr;
+	struct sink_forward_target *sft = appctx->svcctx;
 	struct ring *ring = sink->ctx.ring;
 	struct buffer *buf = &ring->buf;
 	uint64_t msg_len;
-	size_t len, cnt, ofs;
+	size_t len, cnt, ofs, last_ofs;
 	int ret = 0;
 	char *p;
 
+	if (unlikely(se_fl_test(appctx->sedesc, (SE_FL_EOS|SE_FL_ERROR|SE_FL_SHR|SE_FL_SHW))))
+		goto out;
+
 	/* if stopping was requested, close immediately */
 	if (unlikely(stopping))
 		goto close;
 
-	/* for rex because it seems reset to timeout
-	 * and we don't want expire on this case
-	 * with a syslog server
-	 */
-	si_oc(si)->rex = TICK_ETERNITY;
-	/* rto should not change but it seems the case */
-	si_oc(si)->rto = TICK_ETERNITY;
-
-	/* an error was detected */
-	if (unlikely(si_ic(si)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
-		goto close;
-
-	/* con closed by server side */
-	if ((si_oc(si)->flags & CF_SHUTW))
-		goto close;
-
 	/* if the connection is not established, inform the stream that we want
 	 * to be notified whenever the connection completes.
 	 */
-	if (si_opposite(si)->state < SI_ST_EST) {
-		si_cant_get(si);
-		si_rx_conn_blk(si);
-		si_rx_endp_more(si);
-		return;
+	if (sc_opposite(sc)->state < SC_ST_EST) {
+		applet_need_more_data(appctx);
+		se_need_remote_conn(appctx->sedesc);
+		applet_have_more_data(appctx);
+		goto out;
 	}
 
 	HA_SPIN_LOCK(SFT_LOCK, &sft->lock);
@@ -493,7 +477,6 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 		HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 		goto close;
 	}
-	ofs = sft->ofs;
 
 	HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 	LIST_DEL_INIT(&appctx->wait_entry);
@@ -509,99 +492,93 @@ static void sink_forward_oc_io_handler(struct appctx *appctx)
 	 * existing messages before grabbing a reference to a location. This
 	 * value cannot be produced after initialization.
 	 */
-	if (unlikely(ofs == ~0)) {
-		ofs = 0;
-
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
+	if (unlikely(sft->ofs == ~0)) {
+		sft->ofs = b_peek_ofs(buf, 0);
+		HA_ATOMIC_INC(b_orig(buf) + sft->ofs);
 	}
 
+	/* we were already there, adjust the offset to be relative to
+	 * the buffer's head and remove us from the counter.
+	 */
+	ofs = sft->ofs - b_head_ofs(buf);
+	if (sft->ofs < b_head_ofs(buf))
+		ofs += b_size(buf);
+	BUG_ON(ofs >= buf->size);
+	HA_ATOMIC_DEC(b_peek(buf, ofs));
+
 	/* in this loop, ofs always points to the counter byte that precedes
 	 * the message so that we can take our reference there if we have to
 	 * stop before the end (ret=0).
 	 */
-	if (si_opposite(si)->state == SI_ST_EST) {
-		/* we were already there, adjust the offset to be relative to
-		 * the buffer's head and remove us from the counter.
-		 */
-		ofs -= ring->ofs;
-		BUG_ON(ofs >= buf->size);
-		HA_ATOMIC_DEC(b_peek(buf, ofs));
+	ret = 1;
+	while (ofs + 1 < b_data(buf)) {
+		cnt = 1;
+		len = b_peek_varint(buf, ofs + cnt, &msg_len);
+		if (!len)
+			break;
+		cnt += len;
+		BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
 
-		ret = 1;
-		while (ofs + 1 < b_data(buf)) {
-			cnt = 1;
-			len = b_peek_varint(buf, ofs + cnt, &msg_len);
-			if (!len)
-				break;
-			cnt += len;
-			BUG_ON(msg_len + ofs + cnt + 1 > b_data(buf));
-
-			chunk_reset(&trash);
-			p = ulltoa(msg_len, trash.area, b_size(&trash));
-			if (p) {
-				trash.data = (p - trash.area) + 1;
-				*p = ' ';
-			}
-
-			if (!p || (trash.data + msg_len > b_size(&trash))) {
-				/* too large a message to ever fit, let's skip it */
-				ofs += cnt + msg_len;
-				continue;
-			}
-
-			trash.data += b_getblk(buf, p + 1, msg_len, ofs + cnt);
-
-			if (ci_putchk(si_ic(si), &trash) == -1) {
-				si_rx_room_blk(si);
-				ret = 0;
-				break;
-			}
-			ofs += cnt + msg_len;
+		chunk_reset(&trash);
+		p = ulltoa(msg_len, trash.area, b_size(&trash));
+		if (p) {
+			trash.data = (p - trash.area) + 1;
+			*p = ' ';
 		}
 
-		HA_ATOMIC_INC(b_peek(buf, ofs));
-		ofs += ring->ofs;
-		sft->ofs = ofs;
+		if (!p || (trash.data + msg_len > b_size(&trash))) {
+			/* too large a message to ever fit, let's skip it */
+			ofs += cnt + msg_len;
+			continue;
+		}
+
+		trash.data += b_getblk(buf, p + 1, msg_len, ofs + cnt);
+
+		if (applet_putchk(appctx, &trash) == -1) {
+			ret = 0;
+			break;
+		}
+		ofs += cnt + msg_len;
 	}
+
+	HA_ATOMIC_INC(b_peek(buf, ofs));
+	last_ofs = b_tail_ofs(buf);
+	sft->ofs = b_peek_ofs(buf, ofs);
+
 	HA_RWLOCK_RDUNLOCK(LOGSRV_LOCK, &ring->lock);
 
 	if (ret) {
 		/* let's be woken up once new data arrive */
 		HA_RWLOCK_WRLOCK(LOGSRV_LOCK, &ring->lock);
 		LIST_APPEND(&ring->waiters, &appctx->wait_entry);
+		ofs = b_tail_ofs(buf);
 		HA_RWLOCK_WRUNLOCK(LOGSRV_LOCK, &ring->lock);
-		si_rx_endp_done(si);
+		if (ofs != last_ofs) {
+			/* more data was added into the ring between the
+			 * unlock and the lock, and the writer might not
+			 * have seen us. We need to reschedule a read.
+			 */
+			applet_have_more_data(appctx);
+		} else
+			applet_have_no_more_data(appctx);
 	}
 	HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 
+  out:
 	/* always drain data from server */
-	co_skip(si_oc(si), si_oc(si)->output);
+	co_skip(sc_oc(sc), sc_oc(sc)->output);
 	return;
 
 close:
-	si_shutw(si);
-	si_shutr(si);
-	si_ic(si)->flags |= CF_READ_NULL;
+	se_fl_set(appctx->sedesc, SE_FL_EOS|SE_FL_EOI);
+	goto out;
 }
 
 void __sink_forward_session_deinit(struct sink_forward_target *sft)
 {
-	struct stream_interface *si;
-	struct stream *s;
+	struct stream *s = appctx_strm(sft->appctx);
 	struct sink *sink;
 
-	if (!sft->appctx)
-		return;
-
-	si = sft->appctx->owner;
-	if (!si)
-		return;
-
-	s = si_strm(si);
-	if (!s)
-		return;
-
 	sink = strm_fe(s)->parent;
 	if (!sink)
 		return;
@@ -614,10 +591,44 @@ void __sink_forward_session_deinit(struct sink_forward_target *sft)
 	task_wakeup(sink->forward_task, TASK_WOKEN_MSG);
 }
 
+static int sink_forward_session_init(struct appctx *appctx)
+{
+	struct sink_forward_target *sft = appctx->svcctx;
+	struct stream *s;
+	struct sockaddr_storage *addr = NULL;
+
+	if (!sockaddr_alloc(&addr, &sft->srv->addr, sizeof(sft->srv->addr)))
+		goto out_error;
+	/* srv port should be learned from srv->svc_port not from srv->addr */
+	set_host_port(addr, sft->srv->svc_port);
+
+	if (appctx_finalize_startup(appctx, sft->sink->forward_px, &BUF_NULL) == -1)
+		goto out_free_addr;
+
+	s = appctx_strm(appctx);
+	s->scb->dst = addr;
+	s->scb->flags |= (SC_FL_RCV_ONCE|SC_FL_NOLINGER);
+
+	s->target = &sft->srv->obj_type;
+	s->flags = SF_ASSIGNED;
+
+	s->do_log = NULL;
+	s->uniq_id = 0;
+
+	applet_expect_no_data(appctx);
+	sft->appctx = appctx;
+
+	return 0;
+
+ out_free_addr:
+	sockaddr_free(&addr);
+ out_error:
+	return -1;
+}
 
 static void sink_forward_session_release(struct appctx *appctx)
 {
-	struct sink_forward_target *sft = appctx->ctx.sft.ptr;
+	struct sink_forward_target *sft = appctx->svcctx;
 
 	if (!sft)
 		return;
@@ -632,6 +643,7 @@ static struct applet sink_forward_applet = {
 	.obj_type = OBJ_TYPE_APPLET,
 	.name = "<SINKFWD>", /* used for logging */
 	.fct = sink_forward_io_handler,
+	.init = sink_forward_session_init,
 	.release = sink_forward_session_release,
 };
 
@@ -639,68 +651,36 @@ static struct applet sink_forward_oc_applet = {
 	.obj_type = OBJ_TYPE_APPLET,
 	.name = "<SINKFWDOC>", /* used for logging */
 	.fct = sink_forward_oc_io_handler,
+	.init = sink_forward_session_init,
 	.release = sink_forward_session_release,
 };
 
 /*
  * Create a new peer session in assigned state (connect will start automatically)
+ * It sets its context into appctx->svcctx.
  */
 static struct appctx *sink_forward_session_create(struct sink *sink, struct sink_forward_target *sft)
 {
-	struct proxy *p = sink->forward_px;
 	struct appctx *appctx;
-	struct session *sess;
-	struct stream *s;
 	struct applet *applet = &sink_forward_applet;
 
 	if (sft->srv->log_proto == SRV_LOG_PROTO_OCTET_COUNTING)
 		applet = &sink_forward_oc_applet;
 
-	appctx = appctx_new(applet);
+	appctx = appctx_new_here(applet, NULL);
 	if (!appctx)
 		goto out_close;
+	appctx->svcctx = (void *)sft;
 
-	appctx->ctx.sft.ptr = (void *)sft;
-
-	sess = session_new(p, NULL, &appctx->obj_type);
-	if (!sess) {
-		ha_alert("out of memory in peer_session_create().\n");
+	if (appctx_init(appctx) == -1)
 		goto out_free_appctx;
-	}
 
-	if ((s = stream_new(sess, &appctx->obj_type, &BUF_NULL)) == NULL) {
-		ha_alert("Failed to initialize stream in peer_session_create().\n");
-		goto out_free_sess;
-	}
-
-
-	s->target = &sft->srv->obj_type;
-	if (!sockaddr_alloc(&s->si[1].dst, &sft->srv->addr, sizeof(sft->srv->addr)))
-		goto out_free_strm;
-	s->flags = SF_ASSIGNED|SF_ADDR_SET;
-	s->si[1].flags |= SI_FL_NOLINGER;
-
-	s->do_log = NULL;
-	s->uniq_id = 0;
-
-	s->res.flags |= CF_READ_DONTWAIT;
-	/* for rto and rex to eternity to not expire on idle recv:
-	 * We are using a syslog server.
-	 */
-	s->res.rto = TICK_ETERNITY;
-	s->res.rex = TICK_ETERNITY;
-	sft->appctx = appctx;
-	task_wakeup(s->task, TASK_WOKEN_INIT);
+	sft->last_conn = now_ms;
 	return appctx;
 
 	/* Error unrolling */
- out_free_strm:
-	LIST_DELETE(&s->list);
-	pool_free(pool_head_stream, s);
- out_free_sess:
-	session_free(sess);
  out_free_appctx:
-	appctx_free(appctx);
+	appctx_free_on_early_error(appctx);
  out_close:
 	return NULL;
 }
@@ -718,9 +698,21 @@ static struct task *process_sink_forward(struct task * task, void *context, unsi
 	if (!stopping) {
 		while (sft) {
 			HA_SPIN_LOCK(SFT_LOCK, &sft->lock);
-			/* if appctx is NULL, start a new session */
-			if (!sft->appctx)
-				sft->appctx = sink_forward_session_create(sink, sft);
+			/* if appctx is NULL, start a new session
+			 *
+			 * We enforce a tempo to ensure we don't perform more than 1 session
+			 * establishment attempt per second.
+			 */
+			if (!sft->appctx) {
+				uint tempo = sft->last_conn + MS_TO_TICKS(1000);
+
+				if (sft->last_conn == TICK_ETERNITY || tick_is_expired(tempo, now_ms))
+					sft->appctx = sink_forward_session_create(sink, sft);
+				else if (task->expire == TICK_ETERNITY)
+					task->expire = tempo;
+				else
+					task->expire = tick_first(task->expire, tempo);
+			}
 			HA_SPIN_UNLOCK(SFT_LOCK, &sft->lock);
 			sft = sft->next;
 		}
@@ -755,6 +747,54 @@ int sink_init_forward(struct sink *sink)
 	task_wakeup(sink->forward_task, TASK_WOKEN_INIT);
 	return 1;
 }
+
+/* This tries to rotate a file-backed ring, but only if it contains contents.
+ * This way empty rings will not cause backups to be overwritten and it's safe
+ * to reload multiple times. That's only best effort, failures are silently
+ * ignored.
+ */
+void sink_rotate_file_backed_ring(const char *name)
+{
+	struct ring ring;
+	char *oldback;
+	int ret;
+	int fd;
+
+	fd = open(name, O_RDONLY);
+	if (fd < 0)
+		return;
+
+	/* check for contents validity */
+	ret = read(fd, &ring, sizeof(ring));
+	close(fd);
+
+	if (ret != sizeof(ring))
+		goto rotate;
+
+	/* contents are present, we want to keep them => rotate. Note that
+	 * an empty ring buffer has one byte (the marker).
+	 */
+	if (ring.buf.data > 1)
+		goto rotate;
+
+	/* nothing to keep, let's scratch the file and preserve the backup */
+	return;
+
+ rotate:
+	oldback = NULL;
+	memprintf(&oldback, "%s.bak", name);
+	if (oldback) {
+		/* try to rename any possibly existing ring file to
+		 * ".bak" and delete remains of older ones. This will
+		 * ensure we don't wipe useful debug info upon restart.
+		 */
+		unlink(oldback);
+		if (rename(name, oldback) < 0)
+			unlink(oldback);
+		ha_free(&oldback);
+	}
+}
+
 /*
  * Parse "ring" section and create corresponding sink buffer.
  *
@@ -825,10 +865,16 @@ int cfg_parse_ring(const char *file, int linenum, char **args, int kwm)
 			goto err;
 		}
 
+		if (cfg_sink->store) {
+			ha_alert("parsing [%s:%d] : cannot resize an already mapped file, please specify 'size' before 'backing-file'.\n", file, linenum);
+			err_code |= ERR_ALERT | ERR_FATAL;
+			goto err;
+		}
+
 		if (size < cfg_sink->ctx.ring->buf.size) {
 			ha_warning("parsing [%s:%d] : ignoring new size '%llu' that is smaller than current size '%llu' for ring '%s'.\n",
 				   file, linenum, (ullong)size, (ullong)cfg_sink->ctx.ring->buf.size, cfg_sink->name);
-			err_code |= ERR_ALERT | ERR_FATAL;
+			err_code |= ERR_WARN;
 			goto err;
 		}
 
@@ -839,6 +885,65 @@ int cfg_parse_ring(const char *file, int linenum, char **args, int kwm)
 			goto err;
 		}
 	}
+	else if (strcmp(args[0], "backing-file") == 0) {
+		/* This tries to mmap file <file> for size <size> and to use it as a backing store
+		 * for ring <ring>. Existing data are delete. NULL is returned on error.
+		 */
+		const char *backing = args[1];
+		size_t size;
+		void *area;
+		int fd;
+
+		if (!cfg_sink || (cfg_sink->type != SINK_TYPE_BUFFER)) {
+			ha_alert("parsing [%s:%d] : 'backing-file' only usable with existing rings.\n", file, linenum);
+			err_code |= ERR_ALERT | ERR_FATAL;
+			goto err;
+		}
+
+		if (cfg_sink->store) {
+			ha_alert("parsing [%s:%d] : 'backing-file' already specified for ring '%s' (was '%s').\n", file, linenum, cfg_sink->name, cfg_sink->store);
+			err_code |= ERR_ALERT | ERR_FATAL;
+			goto err;
+		}
+
+		/* let's check if the file exists and is not empty. That's the
+		 * only condition under which we'll trigger a rotate, so that
+		 * config checks, reloads, or restarts that don't emit anything
+		 * do not rotate it again.
+		 */
+		sink_rotate_file_backed_ring(backing);
+
+		fd = open(backing, O_RDWR | O_CREAT, 0600);
+		if (fd < 0) {
+			ha_alert("parsing [%s:%d] : cannot open backing-file '%s' for ring '%s': %s.\n", file, linenum, backing, cfg_sink->name, strerror(errno));
+			err_code |= ERR_ALERT | ERR_FATAL;
+			goto err;
+		}
+
+		size = (cfg_sink->ctx.ring->buf.size + 4095UL) & -4096UL;
+		if (ftruncate(fd, size) != 0) {
+			close(fd);
+			ha_alert("parsing [%s:%d] : could not adjust size of backing-file for ring '%s': %s.\n", file, linenum, cfg_sink->name, strerror(errno));
+			err_code |= ERR_ALERT | ERR_FATAL;
+			goto err;
+		}
+
+		area = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
+		if (area == MAP_FAILED) {
+			close(fd);
+			ha_alert("parsing [%s:%d] : failed to use '%s' as a backing file for ring '%s': %s.\n", file, linenum, backing, cfg_sink->name, strerror(errno));
+			err_code |= ERR_ALERT | ERR_FATAL;
+			goto err;
+		}
+
+		/* we don't need the file anymore */
+		close(fd);
+		cfg_sink->store = strdup(backing);
+
+		/* never fails */
+		ring_free(cfg_sink->ctx.ring);
+		cfg_sink->ctx.ring = ring_make_from_area(area, size);
+	}
 	else if (strcmp(args[0],"server") == 0) {
 		if (!cfg_sink || (cfg_sink->type != SINK_TYPE_BUFFER)) {
 			ha_alert("parsing [%s:%d] : unable to create server '%s'.\n", file, linenum, args[1]);
@@ -1013,8 +1118,8 @@ struct sink *sink_new_from_logsrv(struct logsrv *logsrv)
 	/* the servers are linked backwards
 	 * first into proxy
 	 */
-	p->srv = srv;
 	srv->next = p->srv;
+	p->srv = srv;
 
 	/* allocate sink_forward_target descriptor */
 	sft = calloc(1, sizeof(*sft));
@@ -1044,6 +1149,7 @@ struct sink *sink_new_from_logsrv(struct logsrv *logsrv)
 	/* insert into sink_forward_targets
 	 * list into sink
 	 */
+	sft->sink = sink;
 	sft->next = sink->sft;
 	sink->sft = sft;
 
@@ -1064,6 +1170,9 @@ struct sink *sink_new_from_logsrv(struct logsrv *logsrv)
 
 	return sink;
 error:
+	if (srv)
+		srv_drop(srv);
+
 	if (p) {
 		if (p->id)
 			free(p->id);
@@ -1073,22 +1182,11 @@ error:
 		free(p);
 	}
 
-	if (srv) {
-		if (srv->id)
-			free(srv->id);
-		if (srv->conf.file)
-			free((void *)srv->conf.file);
-		if (srv->per_thr)
-		       free(srv->per_thr);
-		free(srv);
-	}
-
 	if (sft)
 		free(sft);
 
 	if (sink) {
-		if (sink->ctx.ring)
-			ring_free(sink->ctx.ring);
+		ring_free(sink->ctx.ring);
 
 		LIST_DELETE(&sink->sink_list);
 		free(sink->name);
@@ -1115,7 +1213,7 @@ int cfg_post_parse_ring()
 			ha_warning("ring '%s' event max length '%u' exceeds size, forced to size '%lu'.\n",
 			           cfg_sink->name, cfg_sink->maxlen, (unsigned long)b_size(&cfg_sink->ctx.ring->buf));
 			cfg_sink->maxlen = b_size(&cfg_sink->ctx.ring->buf);
-			err_code |= ERR_ALERT;
+			err_code |= ERR_WARN;
 		}
 
 		/* prepare forward server descriptors */
@@ -1134,6 +1232,7 @@ int cfg_post_parse_ring()
 				sft->srv = srv;
 				sft->appctx = NULL;
 				sft->ofs = ~0; /* init ring offset */
+				sft->sink = cfg_sink;
 				sft->next = cfg_sink->sft;
 				HA_SPIN_INIT(&sft->lock);
 
@@ -1141,11 +1240,16 @@ int cfg_post_parse_ring()
 				if (!ring_attach(cfg_sink->ctx.ring)) {
 					ha_alert("server '%s' sets too many watchers > 255 on ring '%s'.\n", srv->id, cfg_sink->name);
 					err_code |= ERR_ALERT | ERR_FATAL;
+					ha_free(&sft);
+					break;
 				}
 				cfg_sink->sft = sft;
 				srv = srv->next;
 			}
-			sink_init_forward(cfg_sink);
+			if (sink_init_forward(cfg_sink) == 0) {
+				ha_alert("error when trying to initialize sink buffer forwarding.\n");
+				err_code |= ERR_ALERT | ERR_FATAL;
+			}
 		}
 	}
 	cfg_sink = NULL;
@@ -1274,13 +1378,31 @@ static void sink_init()
 static void sink_deinit()
 {
 	struct sink *sink, *sb;
+	struct sink_forward_target *sft_next;
 
 	list_for_each_entry_safe(sink, sb, &sink_list, sink_list) {
-		if (sink->type == SINK_TYPE_BUFFER)
-			ring_free(sink->ctx.ring);
+		if (sink->type == SINK_TYPE_BUFFER) {
+			if (sink->store) {
+				size_t size = (sink->ctx.ring->buf.size + 4095UL) & -4096UL;
+				void *area = (sink->ctx.ring->buf.area - sizeof(*sink->ctx.ring));
+
+				msync(area, size, MS_SYNC);
+				munmap(area, size);
+				ha_free(&sink->store);
+			}
+			else
+				ring_free(sink->ctx.ring);
+		}
 		LIST_DELETE(&sink->sink_list);
+		task_destroy(sink->forward_task);
+		free_proxy(sink->forward_px);
 		free(sink->name);
 		free(sink->desc);
+		while (sink->sft) {
+			sft_next = sink->sft->next;
+			free(sink->sft);
+			sink->sft = sft_next;
+		}
 		free(sink);
 	}
 }
